@@ -2,6 +2,8 @@
 import argparse
 import configparser
 import csv
+import enum
+import json
 import os
 from collections import namedtuple
 from datetime import datetime, timedelta
@@ -9,6 +11,8 @@ from pathlib import Path
 from typing import List
 
 import matplotlib.pyplot as pyplot
+import requests
+from bs4 import BeautifulSoup
 from github import Github, UnknownObjectException
 
 
@@ -20,6 +24,23 @@ class UnknownRepository(Exception):
     pass
 
 
+class StatKind(enum.Enum):
+    Issues = 0,
+    Stars = 1
+
+
+def write_datetime(obj):
+    if isinstance(obj, datetime):
+        return {"iso": obj.isoformat()}
+
+
+def read_datetime(obj):
+    iso_format = obj.get('iso')
+    if iso_format is not None:
+        return datetime.fromisoformat(iso_format)
+    return obj
+
+
 class ConfigReader:
     ini_file = "ghrepo-stats.ini"
 
@@ -43,11 +64,11 @@ class ConfigReader:
         # first check in the repository root path
         current_path = Path(__file__).parent.parent
         path = current_path / self.ini_file
-        if path.exists:
+        if path.exists():
             return path
         # fall back to home directory
         path = Path.home() / self.ini_file
-        if path.exists:
+        if path.exists():
             return path
         raise FileNotFoundError(
             f"Missing initialization file {self.ini_file}, "
@@ -71,9 +92,14 @@ class ConfigReader:
 
 
 class GitHubStats:
-    def __init__(self, repo_name: str, verbose: bool, csv_file: str = ""):
+    cache_dir = ".ghrepo-stats"
+
+    def __init__(self, repo_name: str, verbose: bool, csv_file: str = "",
+                 show_packages: bool = False, min_stars: int = 0):
         self.repo_name: str = repo_name
         self.verbose = verbose
+        self.show_packages = show_packages
+        self.min_stars = min_stars
         self.config = ConfigReader()
         self.github = Github(self.config.username, self.config.token)
         if csv_file:
@@ -117,14 +143,37 @@ class GitHubStats:
         return self.handle_output(issue_nrs, issue_times, title)
 
     def star_stats(self):
-        stargazers = self.repository().get_stargazers_with_dates()
-        times = []
+        start = datetime.utcnow()
+        stargazers = self.repository().get_stargazers_with_dates().reversed
+        cached, since = self.cached_result(StatKind.Stars)
+        if stargazers.totalCount < len(cached):
+            # void the cache if stars have been removed
+            cached.clear()
+        existing_ids = [e["id"] for e in cached]
+        results = []
         for stargazer in stargazers:
             if self.verbose:
                 print(stargazer.starred_at, stargazer.user.login)
-            times.append(stargazer.starred_at)
+            user_id = stargazer.user.id
+            if user_id in existing_ids:
+                if len(cached) > stargazers.totalCount - len(results):
+                    cached.clear()
+                else:
+                    break
+            results.append({
+                "starred_at": stargazer.starred_at,
+                "id": stargazer.user.id
+            })
+        if results:
+            results.sort(key=lambda s: s["starred_at"])
+            cached.extend(results)
+            self.write_cache(cached, start, StatKind.Stars)
 
-        times = sorted(times)
+        times = []
+        for stargazer in cached:
+            times.append(stargazer["starred_at"])
+        if self.verbose:
+            print(f"Getting stars took {datetime.utcnow() - start}")
         star_nrs = []
         star_times = []
         nr = 0
@@ -165,22 +214,22 @@ class GitHubStats:
         return self.handle_output(commit_size, times, title)
 
     def issue_pr_lifetime(self, show_issues: bool):
-        # count the life time weekly
+        # count the lifetime weekly
         issues = self.collect_issues_or_prs(show_issues)
 
         if not issues:
             return self.handle_output([], [], "")
 
         slots: List[dict] = []
-        start_time = issues[-1].opened
-        now = datetime.now()
+        start_time = issues[0].opened
+        now = datetime.utcnow()
         slot_time = start_time
         while slot_time < now:
             slots.append({"time": slot_time, "nr": 0, "days": 0})
             slot_time = slot_time + timedelta(days=7)
 
         slot_len = len(slots)
-        for issue in reversed(issues):
+        for issue in issues:
             days_from_start = (issue.opened - start_time).days
             index = days_from_start // 7
             days = 7 - days_from_start % 7
@@ -209,22 +258,52 @@ class GitHubStats:
         return self.handle_output(issue_nrs, issue_times, title)
 
     def collect_issues_or_prs(self, collect_issues):
-        results = self.repository().get_issues(state="all")
+        cached, since = self.cached_result(StatKind.Issues)
+        last_number = 0
+        kwargs = {"state": "all"}
+        if cached:
+            kwargs["since"] = since + timedelta(seconds=1)
+            last_number = cached[-1]["number"]
+        start = datetime.utcnow()
+        results = self.repository().get_issues(**kwargs)
+        new_results = []
+        for result in results:
+            new_results.append({
+                "is_pr": result.pull_request is not None,
+                "created_at": result.created_at,
+                "closed_at": result.closed_at,
+                "number": result.number,
+                "state": result.state
+            })
+            if result.number <= last_number:
+                # an existing issues has been closed or reopened -
+                # remove it from cache
+                for c in cached:
+                    if c["number"] == result.number:
+                        cached.remove(c)
+                        break
+        if self.verbose:
+            print(f"Getting issues/prs took {datetime.utcnow() - start}")
+        if new_results:
+            cached.extend(new_results)
+            cached.sort(key=lambda v: v["number"])
+            self.write_cache(cached, start, StatKind.Issues)
+
         Issue = namedtuple("Issue", ["opened", "closed"])
         issues: List[Issue] = []
-        for issue in results:
-            if collect_issues != (issue.pull_request is None):
+        for issue in cached:
+            if collect_issues == issue["is_pr"]:
                 # ignore PRs or issues
                 continue
-            if (issue.closed_at is not None and
-                    (issue.closed_at - issue.created_at).seconds < 60):
+            if (issue["closed_at"] is not None and
+                    (issue["closed_at"] - issue["created_at"]).seconds < 60):
                 # ignore immediately closed issues
                 # happens for imported closed issues
                 continue
             if self.verbose:
-                print(issue.number, issue.created_at, issue.closed_at)
-            issues.append(Issue(opened=issue.created_at,
-                                closed=issue.closed_at))
+                print(issue["number"], issue["created_at"], issue["closed_at"])
+            issues.append(Issue(opened=issue["created_at"],
+                                closed=issue["closed_at"]))
         return issues
 
     def issue_lifetime(self):
@@ -268,6 +347,70 @@ class GitHubStats:
             return False
         return True
 
+    def cache_path(self, stat_kind):
+        org, name = self.repo_name.split("/")
+        return (Path.home() / self.cache_dir / org / name /
+                (stat_kind.name + ".json"))
+
+    def cached_result(self, stat_kind):
+        cache_path = self.cache_path(stat_kind)
+        if cache_path.exists():
+            with open(cache_path) as f:
+                cached = json.load(f, object_hook=read_datetime)
+                return cached["data"], cached["since"]
+        return [], None
+
+    def write_cache(self, result, date, stat_kind):
+        cache_path = self.cache_path(stat_kind)
+        cache_path.parent.mkdir(parents=True, exist_ok=True)
+        cached = {
+            "since": date,
+            "data": result
+        }
+        with open(cache_path, "w") as f:
+            json.dump(cached, f, default=write_datetime)
+
+    def dependents(self):
+        # partly taken from https://stackoverflow.com/a/58772379/12480730
+        subquery = "?dependent_type=PACKAGE" if self.show_packages else ""
+        url = f"https://github.com/{self.repo_name}/network/dependents{subquery}"
+        repos = []
+        while True:
+            r = requests.get(url)
+            soup = BeautifulSoup(r.content, "html.parser")
+            data = [
+                ("{}/{}".format(
+                    t.find('a',
+                           {"data-repository-hovercards-enabled": ""}).text,
+                    t.find('a', {"data-hovercard-type": "repository"}).text
+                ), int(t.div.span.text.strip().replace(",", "")))
+                for t in
+                soup.findAll("div", {"data-test-id": "dg-repo-pkg-dependent"})
+            ]
+            if not data:
+                break
+            repos.extend([r for r in data if r[1] >= self.min_stars])
+
+            button_anchors = soup.find(
+                "div", {"class": "paginate-container"}
+            ).find_all("a")
+            for anchor in button_anchors:
+                if anchor.text == "Next":
+                    url = anchor["href"]
+                    break
+            else:
+                break
+        # sort descending by number of stargazers
+        repos.sort(key=lambda d: d[1], reverse=True)
+        if self.csv_file:
+            names = [r[0] for r in repos]
+            stars = [r[1] for r in repos]
+            self.write_csv(stars, names)
+        else:
+            # if no file was given, just write to stdout
+            for name, star in repos:
+                print(f"{name}\t{star}")
+
 
 def main():
     commands = {
@@ -277,7 +420,8 @@ def main():
         "commits": GitHubStats.commit_stats,
         "codesize": GitHubStats.code_size_change,
         "issue-life": GitHubStats.issue_lifetime,
-        "pr-life": GitHubStats.pr_lifetime
+        "pr-life": GitHubStats.pr_lifetime,
+        "dependents": GitHubStats.dependents
     }
     command_string = ", ".join([f"'{cmd}'" for cmd in commands])
     parser = argparse.ArgumentParser(
@@ -292,6 +436,13 @@ def main():
                         help="Outputs diagnostic information")
     parser.add_argument("--csv", help="Write the output into a csv file "
                                       "with the given file path")
+    parser.add_argument("--packages", action="store_true",
+                        help="Only for dependents: get dependent packages "
+                             "instead of repositories")
+    parser.add_argument("--min-stars", type=int,
+                        help="Only for dependents: limits the output to "
+                             "dependents with at least the given number of "
+                             "stargazers.")
     args = parser.parse_args()
     repo_name = args.repo_name
     sub_command = args.sub_command.lower()
@@ -301,7 +452,8 @@ def main():
               f"Supported commands: {command_string}")
         return 1
     try:
-        stats = GitHubStats(repo_name, args.verbose, args.csv)
+        stats = GitHubStats(repo_name, args.verbose, args.csv,
+                            args.packages, args.min_stars)
         commands[sub_command](stats)
     except Exception as exc:
         print(exc)
